import fastapi as fa
import jinja2
from starlette import responses
from ebookwords import core


environment = jinja2.Environment(
    loader=jinja2.PackageLoader("ebookwords", "interface/web/ebook")
)
router = fa.APIRouter()


@router.post("/upload", response_class=responses.RedirectResponse)
def upload_file(
    uploaded_files: list[fa.UploadFile] = fa.File(),
    book_manager: core.BookManager = fa.Depends(core.get_book_manager),
):
    for upload_file in uploaded_files:
        book_manager.store_ebook(upload_file.file)
    return responses.RedirectResponse(url="/", status_code=303)


@router.get("/", response_class=responses.HTMLResponse)
def list_ebooks(
    book_manager: core.BookManager = fa.Depends(core.get_book_manager),
):
    ebooks = book_manager.db_client.get_book_list()
    template = environment.get_template("ebook_list.html")
    html = template.render(ebooks=ebooks)
    return html


@router.get("/{ebook_id}/", response_class=responses.HTMLResponse)
def list_chapters(
    ebook_id: int,
    book_manager: core.BookManager = fa.Depends(core.get_book_manager),
):
    ebook = book_manager.db_client.get_book(ebook_id)
    if ebook is None:
        raise Exception()

    chapters = book_manager.db_client.get_chapter_list(ebook_id)
    template = environment.get_template("chapter_list.html")
    html = template.render(chapters=chapters)
    return html


@router.get("/{ebook_id}/{chapter_number}/", response_class=responses.HTMLResponse)
def show_chapter(
    ebook_id: int,
    chapter_number: int,
    book_manager: core.BookManager = fa.Depends(core.get_book_manager),
):
    paragraphs = book_manager.db_client.get_chapter_paragraphs(ebook_id, chapter_number)
    template = environment.get_template("chapter.html")
    html = template.render(title="", paragraphs=paragraphs)
    return html



import typing
from ebookwords.adapter.db.models import Book, Chapter, BookRecord, ChapterRecord

__all__ = ["Book", "Chapter", "DatabaseClient", "get_database_client"]


class DatabaseClient(typing.Protocol):
    def add_book(self, ebook: Book) -> None: ...
    def get_book_list(self) -> typing.Iterator[BookRecord]: ...
    def get_chapter_list(self, ebook_id: int) -> typing.Iterator[ChapterRecord]: ...
    def get_book(self, ebook_id: int) -> BookRecord | None: ...
    def get_chapter_paragraphs(
        self, ebook_id: int, chapter_number: int
    ) -> typing.Iterator[str]: ...


def get_database_client() -> DatabaseClient:
    from ebookwords.adapter.db import sql

    return sql.SqlDatabaseClient(".test/test.sqlite")



import datetime
import typing
import sqlmodel as sql
from ebookwords.adapter.db import models


class SqlEbook(sql.SQLModel, table=True):
    __tablename__ = "ebook"  # type: ignore

    id: int | None = sql.Field(default=None, primary_key=True)
    date_added: datetime.datetime = sql.Field(default_factory=datetime.datetime.now)
    title: str


class SqlChapter(sql.SQLModel, table=True):
    __tablename__ = "chapter"  # type: ignore

    id: int | None = sql.Field(default=None, primary_key=True)
    ebook_id: int | None = sql.Field(default=None, foreign_key="ebook.id")
    chapter_number: int
    title: str


class SqlParagraph(sql.SQLModel, table=True):
    __tablename__ = "paragraph"  # type: ignore

    id: int | None = sql.Field(default=None, primary_key=True)
    chapter_id: int | None = sql.Field(default=None, foreign_key="chapter.id")
    paragraph_number: int
    text: str


class SqlDatabaseClient:
    def __init__(self, database_path: str) -> None:
        database_url = f"sqlite:///{database_path}"
        self.engine = sql.create_engine(database_url, echo=True)
        sql.SQLModel.metadata.create_all(self.engine)

    def add_book(self, ebook: models.Book) -> None:
        with sql.Session(self.engine) as session:
            ebook_item = SqlEbook(title=ebook.title)
            session.add(ebook_item)
            session.commit()

            for i, chapter in enumerate(ebook.chapters):
                chapter_item = SqlChapter(
                    ebook_id=ebook_item.id,
                    chapter_number=i,
                    title=chapter.title,
                )
                session.add(chapter_item)
                session.commit()

                for j, paragraph in enumerate(chapter.paragraphs):
                    paragraph_item = SqlParagraph(
                        chapter_id=chapter_item.id,
                        paragraph_number=j,
                        text=paragraph,
                    )
                    session.add(paragraph_item)
                session.commit()

    def get_book_list(self) -> typing.Iterator[models.BookRecord]:
        with sql.Session(self.engine) as session:
            statement = sql.select(SqlEbook)
            results = session.exec(statement)
            for ebook_item in results:
                yield models.BookRecord(
                    id=ebook_item.id or -1,
                    date_added=ebook_item.date_added,
                    title=ebook_item.title,
                )

    def get_chapter_list(self, ebook_id: int) -> typing.Iterator[models.ChapterRecord]:
        with sql.Session(self.engine) as session:
            statement = sql.select(SqlChapter.chapter_number, SqlChapter.title)
            statement = statement.where(SqlChapter.ebook_id == ebook_id)
            results = session.exec(statement)

            for chapter_number, chapter_title in results:
                yield models.ChapterRecord(
                    chapter_number=chapter_number,
                    title=chapter_title,
                )

    def get_book(self, ebook_id: int) -> models.BookRecord | None:
        with sql.Session(self.engine) as session:
            statement = sql.select(SqlEbook)
            statement = statement.where(SqlEbook.id == ebook_id)
            results = session.exec(statement)
            ebook_item = results.first()
            if not ebook_item:
                return None
            return models.BookRecord(
                id=ebook_id,
                date_added=ebook_item.date_added,
                title=ebook_item.title,
            )

    def get_chapter_paragraphs(
        self, ebook_id: int, chapter_number: int
    ) -> typing.Iterator[str]:
        with sql.Session(self.engine) as session:
            statement = sql.select(SqlParagraph.text)
            statement = statement.where(SqlParagraph.chapter_id == SqlChapter.id)
            statement = statement.where(SqlChapter.ebook_id == ebook_id)
            statement = statement.where(SqlChapter.chapter_number == chapter_number)
            results = session.exec(statement)
            return results



import typing
from ebookwords.core import book_manager as bm
from ebookwords.adapter import db

__all__ = ["BookManager", "get_book_manager"]


class BookManager(typing.Protocol):
    db_client: db.DatabaseClient

    def store_ebook(self, file: typing.BinaryIO) -> None: ...


def get_book_manager() -> BookManager:
    from ebookwords.adapter.db import sql

    return bm.BookManager(db_client=db.get_database_client())



import typing
import datetime


class Chapter(typing.NamedTuple):
    title: str
    paragraphs: typing.Iterable[str]


class Book(typing.NamedTuple):
    title: str
    chapters: typing.Iterable[Chapter]


class BookRecord(typing.NamedTuple):
    id: int
    date_added: datetime.datetime
    title: str


class ChapterRecord(typing.NamedTuple):
    chapter_number: int
    title: str



import typing


import lxml.etree as etree
import nh3


def extract_paragraphs(html: str) -> typing.Iterator[str]:
    html = _clean_html(html)
    parser = etree.HTMLParser()
    element = etree.fromstring(html, parser=parser)

    xpath = ".//p | .//h1 | .//h2 | .//h3 | .//h4 | .//h5 | .//h6"
    for paragraph in element.xpath(xpath):
        if paragraph.text:
            yield paragraph.text


def _clean_html(html) -> str:
    html = nh3.clean(
        html,
        tags={
            "p",
            "h1",
            "h2",
            "h3",
            "h4",
            "h5",
            "h6",
        },
    )
    html = f"<div>{html}</div>"
    return html



from ebookwords.adapter import db
from ebookwords.adapter import ebook_parser as parser
from ebookwords.adapter import ebook_extractor as extractor

from dataclasses import dataclass
import typing


@dataclass
class BookManager:
    db_client: db.DatabaseClient

    def store_ebook(self, file: typing.BinaryIO) -> None:
        ebook = parser.parse_epub(file)
        chapters = self._get_chapters(ebook)
        book = db.Book(ebook.title, chapters)
        self.db_client.add_book(book)

    @staticmethod
    def _get_chapters(ebook: parser.Ebook) -> typing.Iterator[db.Chapter]:
        for c in ebook.chapters:
            paragraphs = extractor.extract_paragraphs_from_html(c.html)
            yield db.Chapter(c.title, paragraphs)



import typing

from pathlib import Path

import lxml.etree as etree

import contextlib
import tempfile
import zipfile

from ebookwords.adapter.ebook_parser import models


def parse(file: typing.BinaryIO) -> models.Ebook:
    with _load(file) as epub_folder:
        return _parse_epub_folder(epub_folder)


@contextlib.contextmanager
def _load(epub_file: typing.BinaryIO) -> typing.Generator[Path, None, None]:
    """Loads a file object into an epub folder."""
    with tempfile.NamedTemporaryFile(suffix="epub") as file:
        file.write(epub_file.read())
        with tempfile.TemporaryDirectory() as folder:
            with zipfile.ZipFile(file.name) as zf:
                zf.extractall(folder)
            yield Path(folder)


def _parse_epub_folder(path: Path) -> models.Ebook:
    """Parses the content of an epub folder."""
    root_file = _get_epub_root_file(path)
    content_folder = root_file.parent

    root = etree.parse(root_file).getroot()

    manifest = root.find("{*}manifest")
    if manifest is None:
        raise Exception("Epub is missing manifest.")

    spine = root.find("{*}spine")
    if spine is None:
        raise Exception("Epub is missing spine.")

    title = root.find("{*}metadata/{*}title")
    if title is None or title.text is None:
        raise Exception()

    chapters = [
        _parse_spine_element(element, manifest, content_folder) for element in spine
    ]

    return models.Ebook(title=title.text, chapters=chapters)


def _get_epub_root_file(epub_folder: Path) -> Path:
    """Finds the root file in an epub folder."""
    container_file = Path(epub_folder) / "META-INF" / "container.xml"
    container_root = etree.parse(container_file).getroot()
    root_file = container_root[0][0].get("full-path")
    if root_file is None:
        raise Exception(f"Epub is missing root file {root_file}.")
    return epub_folder / root_file


def _parse_spine_element(
    element: etree.Element, manifest: etree.Element, content_folder: Path
) -> models.EbookChapter:
    idref = element.get("idref")
    if idref is None:
        raise Exception("Spine element is missing 'idref'.")

    xpath = f"{{*}}item[@id='{idref}']"
    item = manifest.find(xpath)
    if item is None:
        raise Exception(
            f"The spine element with id '{idref}' is missing in the manifest."
        )

    item_path = item.get("href")
    if item_path is None:
        raise Exception(
            f"The manifest element with id '{idref}' is missing  the'href' attribute."
        )

    chapter_path = content_folder / item_path
    parser = etree.HTMLParser()
    chapter_element = etree.parse(chapter_path, parser=parser).getroot()

    title_element = chapter_element.find("{*}head/{*}title")
    if title_element is not None and title_element.text is not None:
        title = title_element.text
    else:
        title = chapter_path.stem

    body = chapter_element.find("{*}body")
    if body is None:
        raise Exception(
            f"The chapter file {chapter_path.name} is missing a body element."
        )
    html = etree.tostring(body, method="html", encoding="unicode")

    return models.EbookChapter(title, html)
